@@ -4,15 +4,13 @@
 # Published under the GNU General Public License
 use strict;
 
-use Parse::Pidl::Test;
+use Test::More tests => 2 * 8;
+use FindBin qw($RealBin);
+use lib "$RealBin/../lib";
+use lib "$RealBin";
+use Util qw(test_samba4_ndr);
 
-my %settings = Parse::Pidl::Test::GetSettings(@ARGV);
-
-$settings{'IDL-Arguments'} = ['--quiet', '--parse', '--parser=ndr_test.c', '--header=ndr_test.h'];
-$settings{'IncludeFiles'} = ['ndr_test.h'];
-$settings{'ExtraFiles'} = ['ndr_test.c'];
-
-Parse::Pidl::Test::test_idl("string-pull-empty", \%settings,
+test_samba4_ndr("string-pull-empty", 
 ' [public] void TestString([in,flag(STR_ASCII|LIBNDR_FLAG_STR_SIZE4)] string data);',
 '
 	uint8_t data[] = { 0x00, 0x00, 0x00, 0x00 };
@@ -31,7 +29,7 @@ Parse::Pidl::Test::test_idl("string-pull-empty", \%settings,
 		return 3;
 ');
 
-Parse::Pidl::Test::test_idl("string-ascii-pull", \%settings,
+test_samba4_ndr("string-ascii-pull", 
 '
 	[public] void TestString([in,flag(STR_ASCII|LIBNDR_FLAG_STR_SIZE4)] string data);
 ',
